@@ -1,23 +1,26 @@
 /*
  * Copyright © 2012 Intel Corporation
  *
- * Permission to use, copy, modify, distribute, and sell this software and its
- * documentation for any purpose is hereby granted without fee, provided that
- * the above copyright notice appear in all copies and that both that copyright
- * notice and this permission notice appear in supporting documentation, and
- * that the name of the copyright holders not be used in advertising or
- * publicity pertaining to distribution of the software without specific,
- * written prior permission.  The copyright holders make no representations
- * about the suitability of this software for any purpose.  It is provided "as
- * is" without express or implied warranty.
+ * Permission is hereby granted, free of charge, to any person obtaining
+ * a copy of this software and associated documentation files (the
+ * "Software"), to deal in the Software without restriction, including
+ * without limitation the rights to use, copy, modify, merge, publish,
+ * distribute, sublicense, and/or sell copies of the Software, and to
+ * permit persons to whom the Software is furnished to do so, subject to
+ * the following conditions:
  *
- * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
- * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
- * EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
- * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
- * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
- * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
- * OF THIS SOFTWARE.
+ * The above copyright notice and this permission notice (including the
+ * next paragraph) shall be included in all copies or substantial
+ * portions of the Software.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
+ * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
+ * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
+ * NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
+ * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
+ * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
+ * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
+ * SOFTWARE.
  */
 
 #include "config.h"
@@ -26,6 +29,7 @@
 #include "wayland-client.h"
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdint.h>
 #include <string.h>
 #include <unistd.h>
 #include <sys/mman.h>
@@ -79,21 +83,17 @@ err_free:
 static int
 shm_pool_resize(struct shm_pool *pool, int size)
 {
-	if (ftruncate(pool->fd, size) < 0)
+	if (os_resize_anonymous_file(pool->fd, size) < 0)
 		return 0;
 
-#ifdef HAVE_POSIX_FALLOCATE
-	errno = posix_fallocate(pool->fd, 0, size);
-	if (errno != 0)
-		return 0;
-#endif
-
 	wl_shm_pool_resize(pool->pool, size);
 
 	munmap(pool->data, pool->size);
 
 	pool->data = mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_SHARED,
 			  pool->fd, 0);
+	if (pool->data == (void *)-1)
+		return 0;
 	pool->size = size;
 
 	return 1;
@@ -188,6 +188,7 @@ wl_cursor_destroy(struct wl_cursor *cursor)
 	for (i = 0; i < cursor->image_count; i++)
 		wl_cursor_image_destroy(cursor->images[i]);
 
+	free(cursor->images);
 	free(cursor->name);
 	free(cursor);
 }
@@ -391,17 +392,15 @@ wl_cursor_theme_load(const char *name, int size, struct wl_shm *shm)
 		name = "default";
 
 	theme->name = strdup(name);
+	if (!theme->name)
+		goto out_error_name;
 	theme->size = size;
 	theme->cursor_count = 0;
 	theme->cursors = NULL;
 
-	theme->pool =
-		shm_pool_create(shm, size * size * 4);
-	if (!theme->pool) {
-		free(theme->name);
-		free(theme);
-		return NULL;
-	}
+	theme->pool = shm_pool_create(shm, size * size * 4);
+	if (!theme->pool)
+		goto out_error_pool;
 
 	xcursor_load_theme(name, size, load_callback, theme);
 
@@ -409,6 +408,12 @@ wl_cursor_theme_load(const char *name, int size, struct wl_shm *shm)
 		load_default_theme(theme);
 
 	return theme;
+
+out_error_pool:
+	free(theme->name);
+out_error_name:
+	free(theme);
+	return NULL;
 }
 
 /** Destroys a cursor theme object
@@ -452,9 +457,61 @@ wl_cursor_theme_get_cursor(struct wl_cursor_theme *theme,
 }
 
 /** Find the frame for a given elapsed time in a cursor animation
+ *  as well as the time left until next cursor change.
  *
  * \param cursor The cursor
- * \param time Elapsed time since the beginning of the animation
+ * \param time Elapsed time in ms since the beginning of the animation
+ * \param duration pointer to uint32_t to store time left for this image or
+ *                 zero if the cursor won't change.
+ *
+ * \return The index of the image that should be displayed for the
+ * given time in the cursor animation.
+ */
+WL_EXPORT int
+wl_cursor_frame_and_duration(struct wl_cursor *_cursor, uint32_t time,
+			     uint32_t *duration)
+{
+	struct cursor *cursor = (struct cursor *) _cursor;
+	uint32_t t;
+	int i;
+
+	if (cursor->cursor.image_count == 1) {
+		if (duration)
+			*duration = 0;
+		return 0;
+	}
+
+	i = 0;
+	t = time % cursor->total_delay;
+
+	/* If there is a 0 delay in the image set then this
+	 * loop breaks on it and we display that cursor until
+	 * time % cursor->total_delay wraps again.
+	 * Since a 0 delay is silly, and we've never actually
+	 * seen one in a cursor file, we haven't bothered to
+	 * "fix" this.
+	 */
+	while (t - cursor->cursor.images[i]->delay < t)
+		t -= cursor->cursor.images[i++]->delay;
+
+	if (!duration)
+		return i;
+
+	/* Make sure we don't accidentally tell the caller this is
+	 * a static cursor image.
+	 */
+	if (t >= cursor->cursor.images[i]->delay)
+		*duration = 1;
+	else
+		*duration = cursor->cursor.images[i]->delay - t;
+
+	return i;
+}
+
+/** Find the frame for a given elapsed time in a cursor animation
+ *
+ * \param cursor The cursor
+ * \param time Elapsed time in ms since the beginning of the animation
  *
  * \return The index of the image that should be displayed for the
  * given time in the cursor animation.
@@ -462,18 +519,5 @@ wl_cursor_theme_get_cursor(struct wl_cursor_theme *theme,
 WL_EXPORT int
 wl_cursor_frame(struct wl_cursor *_cursor, uint32_t time)
 {
-	struct cursor *cursor = (struct cursor *) _cursor;
-	uint32_t t;
-	int i;
-
-	if (cursor->cursor.image_count == 1)
-		return 0;
-
-	i = 0;
-	t = time % cursor->total_delay;
-
-	while (t - cursor->cursor.images[i]->delay < t)
-		t -= cursor->cursor.images[i++]->delay;
-
-	return i;
+	return wl_cursor_frame_and_duration(_cursor, time, NULL);
 }
